import HTTP
import Vapor
import Foundation
import ZIPFoundation

/// Server Gzip middlere:
/// 1. checks if the "Accept-Encoding" header contains "gzip"
/// 2. if so, compresses the body and sets the response header "Content-Encoding" to "gzip",
public struct GzipServerMiddleware: Middleware, ServiceType {
    public static func makeService(for worker: Container) throws -> GzipServerMiddleware {
        return .init()
    }

    private let shouldGzip: (_ request: Request) -> Bool

    /// The `shouldGzip` closure is asked for every request whether that request
    /// should allow response gzipping. Returns `true` always by default.
    public init(shouldGzip: @escaping (_ request: Request) -> Bool = { _ in true }) {
        self.shouldGzip = shouldGzip
    }

    public func respond(to request: Request, chainingTo next: Responder) throws -> EventLoopFuture<Response> {
        let acceptsGzip = request.http.headers[.acceptEncoding].first?.contains("gzip") ?? false
        guard acceptsGzip && shouldGzip(request) else {
            return try next.respond(to: request)
        }

        let response = try next.respond(to: request)
        return response.flatMap { response in
            var headers = response.http.headers
            headers.replaceOrAdd(name: .contentEncoding, value: "gzip")
            return response.http.body.consumeData(on: request).map { data in
                let stream  = HTTPChunkedStream(on: request)
                let bufSize = 4096
                var buffer  = ByteBufferAllocator().buffer(capacity: 16)
                let header  : [UInt8] = [0x1f, 0x8b, 0x08, 0x00]
                let header2 : [UInt8] = [0x00, 0x03]
                buffer.write(bytes: header)
                buffer.write(integer: UInt32(Date().timeIntervalSince1970), endianness: .little)
                buffer.write(bytes: header2)
                var write   = stream.write(.chunk(buffer))
                let crc32   = try Data.compress(size: data.count, bufferSize: bufSize,
                    provider: {(offset, readSize) -> Data in
                        return data.subdata(in: offset..<offset+readSize)
                    },
                    consumer: { data -> Void in
                        var buffer  = ByteBufferAllocator().buffer(capacity: bufSize)
                        buffer.write(bytes: data)
                        write = write.flatMap {
                            return stream.write(.chunk(buffer))
                        }
                    })
                write.always {
                    buffer.clear()
                    buffer.write(integer: crc32, endianness: .little)
                    buffer.write(integer: UInt32(data.count), endianness: .little)
                    stream.write(.chunk(buffer)).always {
                        _ = stream.write(.end)
                    }
                }
                let httpResponse = HTTPResponse(status: response.http.status, headers: headers, body: stream)
                return request.response(http: httpResponse)
            }
        }
    }
}
